'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.Speck = undefined;

var _assign = require('babel-runtime/core-js/object/assign');

var _assign2 = _interopRequireDefault(_assign);

var _lodash = require('lodash.defaults');

var _lodash2 = _interopRequireDefault(_lodash);

var _lodash3 = require('lodash.mapvalues');

var _lodash4 = _interopRequireDefault(_lodash3);

var _lodash5 = require('lodash.zip');

var _lodash6 = _interopRequireDefault(_lodash5);

var _anyPromise = require('any-promise');

var _anyPromise2 = _interopRequireDefault(_anyPromise);

var _deasync = require('deasync');

var _deasync2 = _interopRequireDefault(_deasync);

var _biguintFormat = require('biguint-format');

var _biguintFormat2 = _interopRequireDefault(_biguintFormat);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var Speck = function Speck(options) {
  var _this = this;

  var self = this;

  this.machineId = null;

  this.options = (0, _assign2.default)({}, options);
  (0, _lodash2.default)(this.options, {
    date: Date,
    epoch: 0,
    fieldBits: {
      ph: 0, // placeholder
      ts: 42, // timestamp
      dc: 4, // datacenter
      w: 6, // worker
      s: 12 // sequence
    }
  });

  this.fieldBits = options.fieldBits;
  var totalBits = 0;
  for (var f in this.fieldBits) {
    if (this.fieldBits.hasOwnProperty(f)) {
      totalBits += this.fieldBits[f];
    }
  }
  var totalBytes = Math.ceil(totalBits / 8);
  if (totalBytes !== totalBits >> 3) {
    throw new Error('Sum of field bits must be whole bytes');
  }

  var CoordinatorInstance = this.options.coordinator || null;
  if (CoordinatorInstance && typeof CoordinatorInstance.coordinate !== 'function') {
    throw new Error('Coordinator must implement coordinate method');
  }

  var DateInstance = this.options.date;
  if (typeof DateInstance.now !== 'function') {
    // Mock Date.now when testing
    throw new Error('Date instance must implement now method');
  }

  this.fieldBits.m = this.fieldBits.dc + this.fieldBits.w;
  this.fieldBitMasks = (0, _lodash4.default)(this.fieldBits, function (bits) {
    return Math.pow(2, bits) - 1;
  });

  var itr = function () {
    var sequence = 0;
    var overflow = false;
    var lastTime = DateInstance.now() - _this.options.epoch;
    var waitDelayExponent = 0;

    return function () {
      waitDelayExponent = Math.min(waitDelayExponent, 10);
      var time = DateInstance.now() - _this.options.epoch;

      if (_this.machineId === null) {
        return waitRetry(Math.pow(2, waitDelayExponent++));
      }
      if (time === lastTime) {
        if (overflow) {
          return waitRetry(Math.pow(2, waitDelayExponent++));
        }
        sequence = sequence + 1 & _this.fieldBitMasks.s;
        if (sequence === 0) {
          overflow = true;
          return waitRetry(Math.pow(2, waitDelayExponent++));
        }
      } else {
        overflow = false;
        sequence = 0;
      }
      lastTime = time;
      waitDelayExponent = 0;

      var id = _this.buildId(time, sequence, totalBytes);

      if (CoordinatorInstance && typeof CoordinatorInstance.tryHeartbeat === 'function' && typeof CoordinatorInstance.shouldRunHeartbeat === 'function' && CoordinatorInstance.shouldRunHeartbeat()) {
        process.nextTick(function () {
          CoordinatorInstance.tryHeartbeat().catch(function (err) {
            console.error(err);
          });
        });
      }

      return id;
    };
  }();

  var waitRetry = function waitRetry() {
    var delay = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 1;

    return _anyPromise2.default.delay(delay).then(function () {
      return itr();
    });
  };

  var SpeckIdYield = {
    next: function next() {
      // async Promise ID generation
      return _anyPromise2.default.resolve(itr());
    },
    generate: function generate() {
      var format = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 'dec';
      // synchronous ID generation
      var id = itr();

      if (_anyPromise2.default.resolve(id) == id) {
        // it's a promise, deasync it
        id = (0, _deasync2.default)(function (idPromise, done) {
          return idPromise.then(function (_id) {
            done(null, _id);
          });
        })(id);
      }

      if (format === 'raw') {
        return id;
      } else {
        return (0, _biguintFormat2.default)(id, format);
      }
    }
  };

  this.initCoordination(CoordinatorInstance);

  return SpeckIdYield;
};

Speck.prototype.buildId = function (time, sequence, totalBytes) {
  var ph = 0;
  var machineIdBits = this.fieldBits.dc + this.fieldBits.w;
  var id = new Buffer(8);
  id.fill(0);

  // we have 8 bytes here we need to fill
  var f1 = [this.fieldBits.s, machineIdBits, this.fieldBits.ts, this.fieldBits.ph];
  var f2 = [sequence, this.machineId, time, ph];
  var fields = (0, _lodash6.default)(f1, f2);

  var bits = void 0,
      fieldValue = void 0,
      leftoverBits = 0,
      leftoverFieldValue = 0;
  var byteOffset = totalBytes - 1;

  fields.forEach(function (f) {
    var byteValue = void 0,
        bitsNeeded = void 0,
        bitsNeededMask = void 0;
    bits = f[0];
    fieldValue = f[1];
    while (bits + leftoverBits >= 8) {
      byteValue = 0;
      if (leftoverBits) {
        bitsNeeded = 8 - leftoverBits;
        bitsNeededMask = Math.pow(2, bitsNeeded) - 1;
        byteValue = ((fieldValue & bitsNeededMask) << leftoverBits) + leftoverFieldValue;
        fieldValue = Math.floor(fieldValue / (1 << bitsNeeded)) + (byteValue >> 8);
        bits -= bitsNeeded;
        leftoverFieldValue = 0;
        leftoverBits = 0;
        id.writeUInt8(byteValue & 0xFF, byteOffset--);
      } else {
        id.writeUInt8(fieldValue & 0xFF, byteOffset--);
        fieldValue = Math.floor(fieldValue / (1 << 8));
        bits -= 8;
      }
    }
    leftoverBits = bits;
    leftoverFieldValue = fieldValue;
  });

  return id;
};

Speck.prototype.initCoordination = function (CoordinatorInstance) {
  if (CoordinatorInstance) {
    return CoordinatorInstance.coordinate(this.coordinationUpdated.bind(this), this.coordinationFailed.bind(this), 5000);
  } else if (typeof this.options.machineId !== 'undefined') {
    this.machineId = this.options.machineId & this.fieldBitMasks.m;
  } else if (typeof this.options.datacenterId !== 'undefined' || typeof this.options.workerId !== 'undefined') {
    var datacenterId = typeof this.options.datacenterId !== 'undefined' ? this.options.datacenterId & this.fieldBitMasks.dc : 0;
    var workerId = typeof this.options.workerId !== 'undefined' ? this.options.workerId & this.fieldBitMasks.w : 0;
    this.machineId = datacenterId << this.fieldBits.w | workerId;
  } else {
    this.machineId = 0;
  }
  return Promsie.resolve();
};

Speck.prototype.coordinationUpdated = function (err, coordination) {
  if (coordination.machineId) {
    this.machineId = coordination.machineId;
  } else if (typeof coordination.datacenterId !== 'undefined' && typeof coordination.workerId !== 'undefined') {
    var datacenter = coordination.datacenterId & this.fieldBitMasks.dc;
    var worker = coordination.workerId & this.fieldBitMasks.w;
    this.machineId = datacenter << this.fieldBits.w | worker;
  } else {
    this.machineId = coordination;
  }
};

Speck.prototype.coordinationFailed = function (err) {
  if (this.options.fallback && this.options.fallback.coordinator) {
    var CoordinatorInstance = this.options.fallback.coordinator;
    CoordinatorInstance.coordinate(this.coordinationUpdated.bind(this));
  }
};

exports.default = Speck;
exports.Speck = Speck;
//# sourceMappingURL=Speck.js.map